Title   Common code to pciconf and pcibus.snp and wakeup.snp

;

a_proc                  set_current_header_type
                        pushad
                          mov di,00eh
                          call Read_PCI_Byte
                          mov ch,bl
                          test cx,00780h        ; function>0 or bit 7
                          setnz [current_header.hdr_multifunction]
                          and cl,07fh
                          mov [current_header.hdr_type],cl
                        popad
                        ret
a_endp                  set_current_header_type

;

a_proc                  reset_busnumber_informations
                        pushad
                        push es

                          cld
                          push ds
                          pop es
                          mov di,Offset pci_busnumber_used
                          mov cx,Size   pci_busnumber_used
                          mov al,false
                          rep stosb

                          mov di,Offset pci_busnumber_from
                          mov cx,Size   pci_busnumber_from/2
                          mov ax,0ffffh
                          rep stosw

                          mov di,Offset pci_busnumber_subordina
                          mov cx,Size   pci_busnumber_subordina
                          mov al,0ffh
                          rep stosb

                        pop es
                        popad
                        ret
a_endp                  reset_busnumber_informations

;

; bx=bus/device/function

a_proc                  collect_busnumber_informations
                        pushad

                          movzx si,bh
                          mov pci_busnumber_used[si],true

                          ; check for bridges
                          mov cl,[current_header.hdr_type]
                          cmp cl,1
                          jb @@not_bridge
                          cmp cl,2
                          ja @@not_bridge

                          mov di,018h           ; primary/secondary/subordinate
                          call Read_PCI_DWord
                          shr ecx,8             ; cl=secondary/ch=subordinate

                          cmp cl,bh             ; secondary<=current?
                          jbe @@not_bridge
                          cmp ch,cl             ; subordinate<secondary?
                          jb @@not_bridge

                          movzx si,cl
                          mov pci_busnumber_subordina[si],ch
                          shl si,1
                          mov pci_busnumber_from[si],bx
                          shr si,1
                          movzx di,ch

                          ; do for BUS numbers si..di
                          dec si
  @@set_pci_busnumber_used_loop:
                          inc si
                          mov pci_busnumber_used[si],true
IFDEF IS_PCICONF_SNP
                          call is_to_be_configured
                          setz al
                          and al,cfg_recursive
                          or  cfg_recursive_bus_numbers[si],al
ENDIF ; IS_PCICONF_SNP
                          cmp si,di
                          jne @@set_pci_busnumber_used_loop

  @@not_bridge:

                        popad
                        ret
a_endp                  collect_busnumber_informations

;

; bx=bus/device/0
; ds:do_all_pci_functions_proc=procedure to call

a_proc                  do_all_pci_functions
                        pushad
                        push [current_header]

                          sub di,di
                          call Read_PCI_DWord
                          cmp ecx,-1
                          je @@exit

                          call set_current_header_type

                          ; do call for function 0
                          call [do_all_pci_functions_proc]

                          ; multifunction?
                          cmp [current_header.hdr_multifunction],true
                          jne @@exit

                          inc bx
  @@loop:
                          ; check function presence
                          sub di,di
                          call Read_PCI_DWord
                          inc ecx                       ; -1?
                          jz @@next                     ; yes, invaild
                          dec ecx                       ; 0?
                          jz @@next                     ; yes, invaild

                          call set_current_header_type

                          ; do call for function 1..7
                          call [do_all_pci_functions_proc]
  @@next:
                          inc bx
                          test bx,7                     ; function >7?
                          jnz @@loop

  @@exit:
                        pop [current_header]
                        popad
                        ret
a_endp                  do_all_pci_functions
;

; bx=bus/0/0
; ds:do_all_pci_devices_proc=procedure to call
a_proc                  do_all_pci_devices
                        push do_all_pci_functions_proc
                        push cx

                          mov cx,do_all_pci_devices_proc
                          mov do_all_pci_functions_proc,cx
                          mov cx,32
  @@next_device:
                          call do_all_pci_functions
                          add bx,8
                          loop @@next_device

                          sub bx,32*8

                        pop cx
                        pop do_all_pci_functions_proc
                        ret
a_endp                  do_all_pci_devices
;
IFNDEF NO_RESOURCE_CODE
; ecx=address, all writeable bits set
; -> eax resource size(granularity)
a_proc                  guess_size
                        sub eax,eax             ; size:=0
                        test ecx,ecx
                        jz @@exit

                        mov eax,1
  @@loop_guess_size:
                        test ecx,eax
                        jnz @@exit
                        shl eax,1
                        jmp @@loop_guess_size
  @@exit:
                        ret
a_endp                  guess_size

;
; ecx=current address (I/O, prefetch or enable bits unset)
; bx=bus/device/function
; di=register offset
; dl=0:Memory register
; dl=1:I/O register
; -> eax=resource size
p_proc                  guess_size_advanced

                        local @@org_command_register:Word
                        local @@address_register_index:Word
                        local @@address_register_value:DWord

                        push ecx
                        push di

                        mov @@address_register_index,di
                        call Read_PCI_DWord
                        mov @@address_register_value,ecx
                        mov di,004h
                        call Read_PCI_Word
                        mov @@org_command_register,cx

IFDEF OS2
                        ; has address set?
                        mov ecx,@@address_register_value
                        and ecx,esi
                        jz @@can_not_ask_RM

                       ;mov dl,dl               ; pass I/O or Memory flag
                        call GetResourceSize_known_by_ResourceManager
                        jnz @@exit

  @@can_not_ask_RM:

ENDIF ; OS2
                        ; not known by Resource Manager, look at the Hardware...


                        ; must write, when the register value is zero
                        ; and i am trying to allocate(count,alloc)...
IFDEF IS_PCICONF_SNP
                        cmp allocation_phase,allocation_block
                        jne @@must_write
ENDIF ; IS_PCICONF_SNP

                        ; when enabled, use more exact method
                        cmp allow_BAR_modification,true
                        jne @@no_write
    @@must_write:

                        ; unset I/O and memory enable bits in command register
                        mov di,004h
                        mov cx,@@org_command_register
                        mov al,2
                        sub al,dl  ; -> 01(i/O) / 02(mem)
                        not al
                        and cl,al
                        call Write_PCI_Word

                        ; test writeable bits
                        mov di,@@address_register_index
                        mov ecx,esi         ; never set expansion rom enable bit!
                        call Write_PCI_DWord
                        call Read_PCI_DWord
                        and ecx,esi

                        push ecx
                          mov di,@@address_register_index
                          mov ecx,@@address_register_value
                          call Write_PCI_DWord
                          mov di,004h
                          mov cx,@@org_command_register
                          call Write_PCI_Word
                        pop ecx

    @@no_write:
                        call guess_size     ; -> eax

    @@exit:
                        pop di
                        pop ecx
                        ret
p_endp                  guess_size_advanced
;
; di=register
; ecx=current address (I/O, prefetch or enable bits unset)
; esi=mask (ignore i/O<->memory bits)
; dl=0:Memory
; dl=1:I/O
; -> eax=size
a_proc                  guess_size_including_CIS

                        call guess_size_advanced

                        test eax,eax
                        jz @@exit

                        ; using has_CIS_using_BAR_minimum_resource_size+0100 would give
                        ; 'odd' memory sizes/alignments, so better simply set mimimum 100
                        ; and then double it
                        ; note: PCI spec recommends I/O size <=0100..

                        cmp eax,has_CIS_using_BAR_minimum_resource_size
                        ja @@exit

                        cmp eax,0100h
                        if_b <mov eax,0100h>
  @@again:
                        shl eax,1
                        cmp eax,has_CIS_using_BAR_minimum_resource_size
                        jbe @@again
  @@exit:
                        ret
a_endp                  guess_size_including_CIS
ENDIF ; NO_RESOURCE_CODE
;
; bx=bus/device/function
; -> di=start (0 if none)
;
a_proc                  GetCapabiliesStart
                        push eax
                        push ebx
                        push ecx
                        push edx
                        push esi

                          sub dx,dx             ; where tor read capabilies start from
                          mov al,[current_header.hdr_type]
                          cmp al,0
                          if_e <mov dl,034h>
                          cmp al,1
                          if_e <mov dl,034h>
                          cmp al,2
                          if_e <mov dl,014h>
                          mov di,006h
                          call Read_PCI_Word

                          mov di,0
                          test cl,010h          ; no capabilities
                          jz @@exit

                          test dx,dx            ; unknown header type
                          jz @@exit

                          movzx di,dl           ; read start pointer
                          call Read_PCI_Byte
                          movzx di,cl

  @@exit:


                        pop esi
                        pop edx
                        pop ecx
                        pop ebx
                        pop eax
                        ret
a_endp                  GetCapabiliesStart

;

; di=current pointer
; -> CF=1:invalid
a_proc                  IsCapabiliesPointerValid

                        cmp di,0fch             ; validate address
                        jae @@fail

                        cmp di,040h             ; conflict with standard config regsiers, or end of chain (0)
                        jb @@fail

                        test di,00003h          ; dword aligned
                        clc
                        jz @@exit               ; seems to be ok
  @@fail:
                        stc
  @@exit:
                        ret
a_endp                  IsCapabiliesPointerValid

;
; dl=capability
; bx=bus/device/function
; -> di (=0:not found)
a_proc                  SearchCapability
                        push eax
                        push ebx
                        push ecx
                        push edx
                        push esi
                          call GetCapabiliesStart
  @@search_loop:
                          call IsCapabiliesPointerValid
                          jc @@fail

                          call Read_PCI_Word
                          cmp dl,cl
                          je @@exit             ; found!

                          movzx di,ch
                          jmp @@search_loop
  @@fail:
                          sub di,di
  @@exit:
                        pop esi
                        pop edx
                        pop ecx
                        pop ebx
                        pop eax
                        ret
a_endp                  SearchCapability
;
; es:di->es:di
a_proc                  StrEnd
                        dec di
  @@loop:
                        inc di
                        cmp byte Ptr es:[di],0
                        jne @@loop
                        ret
a_endp                  StrEnd
;
IFNDEF NO_RESOURCE_CODE
IFDEF OS2
a_proc                  BlockResources_known_by_ResourceManager
                        mov SearchPurpose_is_block,true
                        jmp BlockResources_known_by_ResourceManager_common
a_endp                  BlockResources_known_by_ResourceManager
;
a_proc                  BlockResources_known_by_ResourceManager_common
                        pushad

                          ; find in the detected trees

                          ; better not reserve for hardware that may be no longer in the machime
                         ;mov eax,HANDLE_PREVIOUS_DETECTED
                         ;call BlockResources_known_by_ResourceManager_searchtree

                          mov eax,HANDLE_CURRENT_DETECTED
                          call BlockResources_known_by_ResourceManager_searchtree


                          cmp SearchPurpose_is_block,true
                          je @@continue_block

                          ; purpose is search - already successful?
                          cmp Search_ResourceSize_know_already,true
                          je @@exit


  @@continue_block:
                          ; find in physical tree

                          ; RMVIEW.EXE->RMINFO.DLL uses an FUNC_RM_ENUM_NODES IOCTL.
                          ; That would helpful to get the device/adapter handles fast..
                          ; Unfortunaly it is not available via the IDC entry

                          call BlockResources_known_by_ResourceManager_PhysTree
  @@exit:


                        popad
                        ret
a_endp                  BlockResources_known_by_ResourceManager_common
;
; searches only in detected trees
; eax=HANDLE_*_DETECTED
a_proc                  BlockResources_known_by_ResourceManager_searchtree
                        push bx

                          mov bl,RM_IDTYPE_RM
                          call BlockResources_known_by_ResourceManager_searchtree_IDType

                          mov bl,RM_IDTYPE_EISA
                          call BlockResources_known_by_ResourceManager_searchtree_IDType

                          mov bl,RM_IDTYPE_PCI
                          call BlockResources_known_by_ResourceManager_searchtree_IDType

                          mov bl,RM_IDTYPE_LEGACY
                          call BlockResources_known_by_ResourceManager_searchtree_IDType

                        pop bx
                        ret
a_endp                  BlockResources_known_by_ResourceManager_searchtree
;
; eax=HANDLE_*_DETECTED
; bl=RM_IDTYPE_*
a_proc                  BlockResources_known_by_ResourceManager_searchtree_IDType
                        push es
                        pushad

                          movzx ebx,bl

                          mov di,Offset d_a_handlelist
                          mov [di].cHandles,0

                          push ds               ; pHndList
                          push di
                          push eax              ; hStartNode
                          push DWord Ptr 0      ; SearchFlags   - no SEARCH_ID_* bits set
                          push DWord Ptr 0      ; SerialNumber  - ignored
                          push DWord Ptr 0      ; VendorID      - ignored
                          push DWord Ptr 0      ; CompatibleID  - ignored
                          push DWord Ptr 0      ; FunctionID    - ignored
                          push DWord Ptr 0      ; DeviceID      - ignored
                          push ebx              ; IDType

                          CallRM RMDevIDToHandleList
                          add sp,9*4
                          test ax,ax
                          jnz @@exit

                          call BlockResources_known_by_ResourceManager_process_d_a_handle_list

  @@exit:
                        popad
                        pop es
                        ret
a_endp                  BlockResources_known_by_ResourceManager_searchtree_IDType
;
; searches only in physical tree
a_proc                  BlockResources_known_by_ResourceManager_PhysTree
                        pushad

                          mov di,Offset d_a_handlelist
                          mov [di].cHandles,0

                          push ds               ; @HandleList
                          push di
                          push ds               ; @SearchKey
                          push Offset SearchKeyAll
                          push HANDLE_PHYS_TREE ; hStartNode
                          CallRM RMKeyToHandleList
                          add sp,4+4+4
                          test ax,ax
                          jnz @@exit

                          call BlockResources_known_by_ResourceManager_process_d_a_handle_list

  @@exit:
                        popad
                        ret
a_endp                  BlockResources_known_by_ResourceManager_PhysTree
;
a_proc                  BlockResources_known_by_ResourceManager_process_d_a_handle_list
                        pushad

                          mov di,Offset d_a_handlelist

                          ; look at all devices/adapters for resources

                          mov cx,[di].cHandles
                          test cx,cx
                          jz @@exit

                          lea si,[di].Handles_array
  @@loop:
                          mov eax,[si]
                          add si,4
                          call BlockResources_known_by_ResourceManager_process_d_a_handle
                          loop @@loop
  @@exit:
                        popad
                        ret
a_endp                  BlockResources_known_by_ResourceManager_process_d_a_handle_list
;
; eax=single device/adapter handle
a_proc                  BlockResources_known_by_ResourceManager_process_d_a_handle
                        push es
                        pushad

                          ; get node information from handle - includes resources

                          push Size NodeInfo    ; NodeInfoSize
                          push ds               ; @NodeInfo
                          push Offset NodeInfo
                          push eax              ; hRMHandle
                          CallRM RMGetNodeInfo
                          add sp,2+4+4
                          test ax,ax
                          jnz @@exit


                          ; get PCI address from Adjunct, depending on
                          ; record type...
                          mov bx,0ffffh         ; PCI address (Adjunct format)
                          mov si,Offset NodeInfo.RMNode
                          mov eax,[si].NodeType
                          les si,[si].pNodeData
                          cmp eax,RMTYPE_ADAPTER
                          jne @@not_adapter
                          lea di,[si].pAdjunctList_AS
                          call search_ADJ_PCI_DEVFUNC
  @@not_adapter:
                          cmp eax,RMTYPE_DEVICE
                          jne @@not_device
                          lea di,[si].pAdjunctList_DS
                          call search_ADJ_PCI_DEVFUNC
  @@not_device:
                          cmp eax,RMTYPE_LDEV
                          jne @@not_ldevice
                          lea di,[si].pAdjunctList_LD
                          call search_ADJ_PCI_DEVFUNC
  @@not_ldevice:
                          cmp eax,RMTYPE_SYSNAME
                          jne @@not_sysname
                          lea di,[si].pAdjunctList_SNS
                          call search_ADJ_PCI_DEVFUNC
  @@not_sysname:
                          cmp eax,RMTYPE_DETECTED
                          jne @@not_detected
                          lea di,[si].pAdjunctList_DTS
                          call search_ADJ_PCI_DEVFUNC
  @@not_detected:
                          ; other types: RMTYPE_DRIVER,RMTYPE_RESOURCE, do not have pAdjunctList



                          ; has resources?
                          mov si,Offset NodeInfo.RMNode
                          cmp [si].pResourceList,0
                          je @@no_resource_list

                          les si,[si].pResourceList
                          mov edi,es:[si].Count_RS      ; use edi as countdown variable
                          add si,Offset Resource_Array
                          test edi,edi
                          jz @@done_resource_list
  @@loop_resource_list:
                          ; es:si=@ResourceStruct
                          cmp es:[si].ResourceType,RS_TYPE_IO
                          jne @@not_resource_IO

                          ; ..IOResource
                          mov edx,1
                          mov cx,es:[si].resource_record.IOAddressLines
                          cmp cx,10             ; protect from invalid values..
                          if_b <mov cx,10>
                          cmp cx,16
                          if_a <mov cx,16>
                          shl edx,cl
                          ; there are some BIOS that return bad pnp data,
                          ; for example serial ports every 400h, even on
                          ; pci chipsets (10 instead of 10h problem suspeced..)
                          ; we could add special hacks for such errors later,
                          ; for now be pedantic..
                          movzx ecx,es:[si].resource_record.BaseIOPort
                          movzx eax,es:[si].resource_record.NumIOPorts
                          dec edx
                          and ecx,edx
                          inc edx
                          ; IOFlags is ignored
  @@io_alias_loop:
                          call FoundIORange
                          add ecx,edx           ; for 16 bit, this wraps
                          test ecx,0ffff0000h
                          jz  @@io_alias_loop   ; for 10 bit, it may take some iterations..
                          jmp @@done_resource
  @@not_resource_IO:
                          cmp es:[si].ResourceType,RS_TYPE_MEM
                          jne @@not_resource_MEM
                          ; ..MemResource
                          mov ecx,es:[si].resource_record.MemBase
                          mov eax,es:[si].resource_record.MemSize
                          ; MemFlags is ignored
                          call FoundMemoryRange
                         ;jmp @@done_resource
  @@not_resource_MEM:
  @@done_resource:
                          add si,Size ResourceStruct
                          dec edi
                          jnz @@loop_resource_list
  @@done_resource_list:
  @@no_resource_list:
  @@exit:

                        popad
                        pop es
                        ret
a_endp                  BlockResources_known_by_ResourceManager_process_d_a_handle
;
; es:di=@pAdjunctList
; -> bx=PCI_DEVFUNC value
a_proc                  search_ADJ_PCI_DEVFUNC
                        push es
                        push di

                          ; no (more) Adjunct data?
  @@loop:
                          cmp es:[di].pNextAdj,0
                          je @@exit

                          les di,es:[di].pNextAdj
                          cmp es:[di].AdjType,ADJ_PCI_DEVFUNC
                          if_e <mov bx,Word Ptr es:[di+Size AdjHeader]>

                          jmp @@loop

  @@exit:
                        pop di
                        pop es
                        ret
a_endp                  search_ADJ_PCI_DEVFUNC
;
; ecx..ecx+eax-1
; bx=bus/device/function
a_proc                  FoundIORange

IFDEF IS_PCICONF_SNP
                        cmp SearchPurpose_is_block,true
                        je BlockIORange
ENDIF ; IS_PCICONF_SNP
                        cmp Search_ResourceType_IO,1
                        je FoundResourceRange

                        ret
a_endp                  FoundIORange
;
; ecx..ecx+eax-1
; bx=bus/device/function
a_proc                  FoundMemoryRange

IFDEF IS_PCICONF_SNP
                        cmp SearchPurpose_is_block,true
                        je BlockMemoryRange
ENDIF ; IS_PCICONF_SNP
                        cmp Search_ResourceType_IO,0
                        je FoundResourceRange

                        ret
a_endp                  FoundMemoryRange
;
a_proc                  FoundResourceRange

                        ; start address matches?
                        cmp Search_ResourceAddress,ecx
                        jne @@exit

                        ; already know better?
                        cmp Search_ResourceSize_know_already,true
                        je @@exit

                        ; when having the size detected by PCI snoop, assume that
                        ; this size is the best value to use, if multiple
                        ; drivers have registered a resource
                        ; when the resource is assigned to an differnt PCI device,
                        ; assume this is an error
                        cmp bx,0ffffh                   ; non-pci is accepted
                        je @@not_exact
                        cmp Search_pci_addr,bx          ; wrong device?
                        jne @@exit

                        mov Search_ResourceSize_know_already,true ; PCIBUS.SNP did it, assume very good value..
  @@not_exact:
                        mov Search_ResourceSize,eax
  @@exit:
                        ret
a_endp                  FoundResourceRange
;
; ecx=address
; bx=bus/device/function
; dl=resource type: 0=Memory 1=I/O
; -> ZF=0:success eax=size
;    ZF=1:error   eax=0
a_proc                  GetResourceSize_known_by_ResourceManager
                        ; fast way would be RMAdjToHandleList(HANDLE_CURRENT_DETECTED) for bus/device/function adjunct
                        ; but HANDLE_CURRENT_DETECTED causes RMRC_BAD_DEVICEHANDLE, so use the above code

                        mov SearchPurpose_is_block,false
                        Set_PCI_DevFunc Search_pci_addr,bx
                        mov Search_ResourceType_IO,dl
                        mov Search_ResourceAddress,ecx
                        mov Search_ResourceSize,0    ; invalid size
                        mov Search_ResourceSize_know_already,false

                        ; should not try to find 0 (DMA controller for IO or 640KiB memory)
                        cmp Search_ResourceAddress,0
                        je @@search_failed

                        call BlockResources_known_by_ResourceManager_common

  @@search_failed:
                        mov eax,Search_ResourceSize
                        test eax,eax

                        ret
a_endp                  GetResourceSize_known_by_ResourceManager
;
ENDIF ; OS2
ENDIF ; NO_RESOURCE_CODE

